'''
Ctypes interface for struct-based interface to the C-version of NRLMSISE-00.
This C version of NRLMSISE-00 is written by Dominik Brodowski 
'''

from ctypes import (cdll, Structure, c_int, c_double, pointer, byref, POINTER)
import os
base = os.path.dirname(os.path.abspath(__file__))

for fn in os.listdir(base):
    if 'libnrlmsis' in fn and (fn.endswith('.so') or
                               fn.endswith('.dll') or 
                               fn.endswith('.dylib')):
        msis = cdll.LoadLibrary(os.path.join(base, fn))
        break


class nrlmsise_flags(Structure):
    """C-struct containing NRLMSISE related switches"""
    _fields_ = [("switches", c_int * 24), ("sw", c_double * 24),
                ("swc", c_double * 24)]


class ap_array(Structure):
    """C-struct containing NRLMSISE related switches"""
    _fields_ = [("a", c_double * 7)]


class nrlmsise_input(Structure):
    """The C-struct contains input variables for NRLMSISE."""
    _field_ = [("year", c_int), ("doy", c_int), ("sec", c_double),
               ("alt", c_double), ("g_lat", c_double), ("g_long", c_double),
               ("lst", c_double), ("f107A", c_double), ("f107", c_double),
               ("ap", c_double), ("ap_a", POINTER(ap_array))]


class nrlmsise_output(Structure):
    """The C-struct contains output variables for NRLMSISE."""
    _fields_ = [("d", c_double * 9), ("t", c_double * 2)]



from MCEq.misc import info
import six
import MCEq.geometry.nrlmsise00.nrlmsise00 as cmsis


class NRLMSISE00Base(object):
    def __init__(self):
        # Cache altitude value of last call
        self.last_alt = None

        self.inp = cmsis.nrlmsise_input()
        self.output = cmsis.nrlmsise_output()
        self.flags = cmsis.nrlmsise_flags()
        self.month2doy = {
            'January': 1,
            'February': 32,
            'March': 60,
            'April': 91,
            'May': 121,
            'June': 152,
            'July': 182,
            'August': 213,
            'September': 244,
            'October': 274,
            'November': 305,
            'December': 335
        }
        # Longitude, latitude, height
        self.locations = {
            'SouthPole': (0., -90., 2834. * 100.),
            'Karlsruhe': (8.4, 49., 110. * 100.),
            'Geneva': (6.1, 46.2, 370. * 100.),
            'Tokyo': (139., 35., 5. * 100.),
            'SanGrasso': (13.5, 42.4, 5. * 100.),
            'TelAviv': (34.8, 32.1, 5. * 100.),
            'KSC': (-80.7, 32.1, 5. * 100.),
            'SoudanMine': (-92.2, 47.8, 5. * 100.),
            'Tsukuba': (140.1, 36.2, 30. * 100.),
            'LynnLake': (-101.1, 56.9, 360. * 100.),
            'PeaceRiver': (-117.2, 56.15, 36000. * 100.),
            'FtSumner': (-104.2, 34.5, 31000. * 100.)
        }

        self.daytimes = {'day': 43200., 'night': 0.}
        self.current_location = 'SouthPole'
        self.init_default_values()

    def surface_vert_depth(self, loc='SouthPole', month='June'):
        self.set_location('SouthPole')
        self.set_season('June')

    def height2depth(self, altitude_cm):
        from scipy.integrate import quad
        return quad(self.get_density, altitude_cm, 112.8 * 1e5,
                    epsrel=0.001)[0]

    def _retrieve_result(self, *args, **kwargs):
        """Calls NRLMSISE library's main function"""
        raise Exception('Not implemented for the base class')

    def get_temperature(self, altitude_cm):
        """Returns temperature in K"""
        self._retrieve_result(altitude_cm)
        return self.output.t[1]

    def get_density(self, altitude_cm):
        """Returns density in g/cm^3"""
        self._retrieve_result(altitude_cm)
        return self.output.d[5]


class cNRLMSISE00(NRLMSISE00Base):
    def init_default_values(self):
        """Sets default to June at South Pole"""

        self.inp.doy = cmsis.c_int(self.month2doy['June'])  # Day of year
        self.inp.year = cmsis.c_int(0)  # No effect
        self.inp.sec = cmsis.c_double(self.daytimes['day'])  # 12:00
        self.inp.alt = cmsis.c_double(self.locations[self.current_location][2])
        self.inp.g_lat = cmsis.c_double(
            self.locations[self.current_location][1])
        self.inp.g_long = cmsis.c_double(
            self.locations[self.current_location][0])
        self.inp.lst = cmsis.c_double(self.inp.sec.value / 3600. +
                                      self.inp.g_long.value / 15.)
        # Do not touch this except you know what you are doing
        self.inp.f107A = cmsis.c_double(150.)
        self.inp.f107 = cmsis.c_double(150.)
        self.inp.ap = cmsis.c_double(4.)
        self.inp.ap_a = cmsis.pointer(cmsis.ap_array())
        self.alt_surface = self.locations[self.current_location][2]

        self.flags.switches[0] = cmsis.c_int(0)
        for i in range(1, 24):
            self.flags.switches[i] = cmsis.c_int(1)

    def set_location(self, tag):
        if tag not in list(self.locations):
            raise Exception(
                "NRLMSISE00::set_location(): Unknown location tag '{0}'.".
                format(tag))

        self.inp.alt = cmsis.c_double(self.locations[tag][2])
        self.set_location_coord(*self.locations[tag][:2])
        self.current_location = tag
        self.alt_surface = self.locations[self.current_location][2]

    def set_location_coord(self, longitude, latitude):
        info(5, 'long={0:5.2f}, lat={1:5.2f}'.format(longitude, latitude))
        if abs(latitude) > 90 or abs(longitude) > 180:
            raise Exception("NRLMSISE00::set_location_coord(): Invalid inp.")
        self.inp.g_lat = cmsis.c_double(latitude)
        self.inp.g_long = cmsis.c_double(longitude)

    def set_season(self, tag):
        if tag not in self.month2doy:
            raise Exception("NRLMSISE00::set_location(): Unknown season tag.")
        info(5, 'Season', tag, 'doy=', self.month2doy[tag])
        self.inp.doy = self.month2doy[tag]

    def set_doy(self, doy):
        if doy < 0 or doy > 365:
            raise Exception("NRLMSISE00::set_doy(): Day of year out of range.")
        info(5, 'day of year', doy)
        self.inp.doy = doy

    def _retrieve_result(self, altitude_cm):
        if self.last_alt == altitude_cm:
            return

        inp = self.inp
        inp.alt = cmsis.c_double(altitude_cm / 1e5)
        cmsis.msis.gtd7_py(inp.year, inp.doy, inp.sec, inp.alt, inp.g_lat,
                           inp.g_long, inp.lst, inp.f107A, inp.f107, inp.ap,
                           inp.ap_a, cmsis.byref(self.flags),
                           cmsis.byref(self.output))

        self.last_alt = altitude_cm


def test():
    import numpy as np
    import matplotlib.pyplot as plt

    msis = cNRLMSISE00()
    den = np.vectorize(msis.get_density)

    plt.figure(figsize=(16, 5))
    plt.suptitle('NRLMSISE-00')

    h_vec = np.linspace(0, 112.8 * 1e5, 500)
    msis.set_season('January')
    msis.set_location('SouthPole')
    den_sp_jan = den(h_vec)

    msis.set_season('January')
    msis.set_location('Karlsruhe')
    den_ka_jan = den(h_vec)

    plt.subplot(131)
    plt.semilogy(h_vec / 1e5, den_sp_jan, label='MSIS00: SP Jan.')
    plt.semilogy(h_vec / 1e5, den_ka_jan, label='MSIS00: KA Jan.')
    plt.legend()
    plt.xlabel('vertical height in km')
    plt.ylabel(r'density $\rho(h)$ in g/cm$^3$')

    plt.subplot(132)
    plt.plot(h_vec / 1e5, den_ka_jan / den_sp_jan, label='MSIS00: KA/SP')
    plt.xlabel('vertical height in km')
    plt.ylabel(r'density ratio')
    plt.legend(loc='upper left')

    plt.subplot(133)
    msis.set_location('SouthPole')
    for i in range(360 / 30):
        msis.inp.doy = i * 30
        plt.plot(h_vec / 1e5, den(h_vec) / den_sp_jan, label=str(i + 1))
    plt.legend(ncol=2, loc=3)
    plt.title('MSIS00: SouthPole')
    plt.xlabel('vertical height in km')
    plt.ylabel(r'$\rho$(Month) / $\rho$(January)')
    plt.ylim(ymin=0.6)
    plt.tight_layout()

    plt.figure(figsize=(6, 5))
    h2d = np.vectorize(msis.height2depth)
    plt.semilogy(h_vec / 1e5, h2d(h_vec))
    plt.ylabel(r'Slant depth X [g/cm$^2$]')
    plt.xlabel(r'Atmospheric height $h$ [km]')
    plt.subplots_adjust(left=0.15, bottom=0.11)
    plt.show()


if __name__ == '__main__':
    test()



import sys
from os.path import join
from setuptools import setup, Extension

from distutils.command import build_ext

def get_export_symbols(self, ext):
    """From https://bugs.python.org/issue35893"""
    parts = ext.name.split(".")
    # print('parts', parts)
    if parts[-1] == "__init__":
        initfunc_name = "PyInit_" + parts[-2]
    else:
        initfunc_name = "PyInit_" + parts[-1]

build_ext.build_ext.get_export_symbols = get_export_symbols


# Require pytest-runner only when running tests
needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
pytest_runner = ['pytest-runner'] if needs_pytest else []

setup_requires = pytest_runner

libnrlmsise00 = Extension(
    'MCEq.geometry.nrlmsise00.libnrlmsise00',
    sources=[
        join('MCEq/geometry/nrlmsise00', sf)
        for sf in ['nrlmsise-00_data.c', 'nrlmsise-00.c']
    ],
    include_dirs=['MCEq/geometry/nrlmsise00'])
    
libcorsikaatm = Extension(
    'MCEq.geometry.corsikaatm.libcorsikaatm',
    sources=['MCEq/geometry/corsikaatm/corsikaatm.c'])


# This method is adopted from iMinuit https://github.com/scikit-hep/iminuit
# Getting the version number at this point is a bit tricky in Python:
# https://packaging.python.org/en/latest/development.html#single-sourcing-the-version-across-setup-py-and-your-project
# This is one of the recommended methods that works in Python 2 and 3:
def get_version():
    version = {}
    with open("MCEq/version.py") as fp:
        exec (fp.read(), version)
    return version['__version__']


__version__ = get_version()

setup(
    name='MCEq',
    version=__version__,
    description='Numerical cascade equation solver',
    author='Anatoli Fedynitch',
    author_email='afedynitch@gmail.com',
    license='BSD 3-Clause License',
    url='https://github.com/afedynitch/MCEq',
    packages=['MCEq', 'MCEq.geometry', 'MCEq.geometry.nrlmsise00',
        'MCEq.geometry.corsikaatm'],
    setup_requires=[] + pytest_runner,
    package_dir={
        'MCEq': 'MCEq',
        'MCEq.geometry': 'MCEq/geometry',
        'MCEq.geometry.nrlmsise00': 'MCEq/geometry/nrlmsise00',
        'MCEq.geometry.corsikaatm': 'MCEq/geometry/corsikaatm'
    },
    package_data={
        'MCEq': ['data/README.md'],
    },
    install_requires=[
        'six',
        'h5py',
        'particletools==1.1.3',
        'crflux==1.0.2',
        'scipy',
        'numpy',
        'tqdm',
        'requests'
    ],
    py_modules=['mceq_config'],
    requires=[
        'numpy', 'scipy', 'particletools', 'crflux', 'h5py'
    ],
    ext_modules=[libnrlmsise00, libcorsikaatm],
    extras_require={
        'MKL': ['mkl==2019.0'],
        'CUDA': ['cupy-cuda101==6.2.0']
    },
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering :: Physics',
        'Intended Audience :: Science/Research',
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License'
    ])
